(* ald.v *)
(* ALD : Memory with Allocation and AT disabled *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.


Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module ALD_State <: Mem_T.

(* alloctbl - shows which pages are allocated *)
Record State := { data : Z -> Z; alloctbl : Z -> bool }.

Definition t := State.

Definition ValidAddr (l : Z) : SCheck (State := State) :=
fun s => s.(alloctbl) (Pg l) && ValidPA l.

Definition getMemCheck (l:Z) : SCheck (State := State) := ValidAddr l.
Definition getMem (l:Z) : SGet (State := State) Z := fun s => s.(data) l.

Definition setMemCheck (l : Z) : SCheck (State := State) := ValidAddr l.
Definition setMem (l z : Z) :  SAct (State := State) :=
  fun s => {| data := (fun l' => if Zeq_bool l l' then z else s.(data) l'); alloctbl := s.(alloctbl) |}.


Definition allocpg (pg : Z) : SAct (State := State) :=
  fun s => {| data := s.(data); 
          alloctbl := fun pg' => if Zeq_bool pg pg' then true else s.(alloctbl) pg' |}.

Definition freepg (pg : Z) : SAct (State := State) :=
  fun s => {| data := s.(data); 
          alloctbl := fun pg' => if Zeq_bool pg pg' then false else s.(alloctbl) pg' |}.


Lemma alloctbl_setMem_unch : forall l v m1 pg,
ALD_State.alloctbl (ALD_State.setMem l v m1) pg = alloctbl m1 pg.
autosolve.
Qed.

(* getMem simplifications *)

Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, getMemCheck, ValidAddr, ValidPA in *; simpl; subst.
destruct s; simpl in *.
f_equal.
repeat autoclean.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

Lemma getMem_setMem : 
  forall l l' z s, getMem l (setMem l' z s) = if Zeq_bool l l' then z else getMem l s.
intros.
unfold getMem, setMem; simpl.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
Qed.

Hint Rewrite getMem_setMem : mem.

End ALD_State. 
