(* ald_lib.v *)
(* The stub library for the ALD memory model *)
(* Contains the specifications of mem_alloc and mem_free *)
(* TODO: move the ALD_pmap_init_spec (a_{ALD}^{ptinit}) somewhere else *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import label.
Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import c.
Require Import ald.

Module ALD_Lib.

Import C_ALD.

Definition allocupd (tbl : Z -> bool) (pg : Z) (b : bool) : Z -> bool :=
   fun x => if Zeq_bool x pg then b else tbl x.

Definition allocfail : SAct (State := State) :=
  StackAct (pop +++ setRet 0%Z).

Definition allocpg (pg : Z) : State -> State -> Prop :=
  fun s s' => 
	 (forall l, Pg l <> pg -> getMem l s'.(memory) = getMem l s.(memory)) /\ 
	 (s'.(memory).(alloctbl) = allocupd s.(memory).(alloctbl) pg true) /\
	s'.(stack) = (pop +++ setRet pg) s.(stack).


Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 0) s = true /\ StackCheck isFrame s = true
     );

rel := (fun s s' =>
    (exists pg, (1 <= pg < NPAGES)%Z /\ s.(memory).(alloctbl) pg = false /\ allocpg pg s s')
    \/
    s' = allocfail s)
|}.

Definition mem_free_spec : Action State :=
SAct_Action (fun s => StackCheck (getArgsCheck 1) s && 
                                 let pg := nth 0 (StackGet getArgs s) 0%Z in
                                 (* ValidPg pg && *) Zlt_bool 0 pg && Zlt_bool pg NPAGES && s.(memory).(alloctbl) pg) 
                   (fun s =>
                            SAct_chain 
                                  (StackAct (pop +++ setRet 0%Z)) 
                                  (MemAct (fun m => {| data := m.(data); 
                                        alloctbl := allocupd m.(alloctbl) (nth 0 (StackGet getArgs s) 0%Z) false |}))
                                   s
                   ).

 
Definition ald_lib_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | _ => None
  end.


Require Import ale.
Require Import pe.

Definition ALD_pmap_init_spec : Action C_ALD.State.t :=
 {| pre := (fun (s : C_ALD.State.t) => C_ALD.State.StackCheck (getArgsCheck 0) s = true /\ 
                                          (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                                                     s.(C_ALD.State.memory).(ALD_State.alloctbl) (Pg l) = true)
                                           ); 
   rel := fun (s s' : C_ALD.State.t) => 
     (forall pg, ValidPg pg = true -> s'.(C_ALD.State.memory).(ALD_State.data) (PTROOT + pg * 8)%Z = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s'.(C_ALD.State.memory).(ALD_State.data) (PTROOT + pg * 8)%Z= 0%Z) /\
     (forall l, ValidPA l = true -> ALE_State.PtDom l = false -> 
             s'.(C_ALD.State.memory).(ALD_State.data) l = s.(C_ALD.State.memory).(ALD_State.data) l) /\
     (forall pg, ValidPg pg = true -> s'.(C_ALD.State.memory).(ALD_State.alloctbl) pg = 
                                                    s.(C_ALD.State.memory).(ALD_State.alloctbl) pg) /\
     s'.(C_ALD.State.stack) = setRet 0 (pop s.(C_ALD.State.stack))
|}. 

End ALD_Lib.
