(* as.v *)
(* Definition of the AS memory model *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module AS_State <: Mem_T.

Record State := { data : Z -> Z; allocdata : Z -> bool}.
Definition t := State.

Definition addrCheck (addr : Z) : SCheck (State := State) := 
fun s => ValidVA addr && s.(allocdata) (Pg addr).

Definition getMemCheck (l:Z) : SCheck (State := State) := 
fun s => addrCheck l s.
 
Definition getMem (l:Z) : SGet (State := State) Z :=
fun s => s.(data) l.

Definition setMemCheck (l : Z) : SCheck (State := State) := 
fun s => addrCheck l s.

Definition setMem (l z : Z) :  SAct (State := State) :=
fun s => {| data := fun l' => if Zeq_bool l l' then z else s.(data) l'; allocdata := s.(allocdata) |}.

(* getMem simplifications *)
Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, ValidVA, ValidPA in *; simpl; subst.
destruct s; simpl in *.
f_equal; simpl.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

Lemma getMem_setMem : 
  forall l l' z s, getMem l (setMem l' z s) = if Zeq_bool l l' then z else getMem l s.
intros.
unfold getMem, setMem; simpl.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
Qed.

Hint Rewrite getMem_setMem : mem.


End AS_State. 
