(* hw_lib.v *)
(* The stub library for the HW memory model (setPE and setPTROOT) . *)
(* Also defines the HW stubs for kernel_init, mem_init, and pt_init *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import label.
Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.
Require Export common.
Require Import sact.
Require Import actsmp.

Require Import c.
Require Import hw.

Module HW_Lib.

Require Import ale.
Require Import pe.

Opaque PE_State.PMM_data_loc.

Import C_HW.

Definition pe_enable_spec : Action State :=
SAct_Action (StackCheck (getArgsCheck 0))
	           (StackAct (pop +++ setRet 0%Z) +++ 
	             (MemAct (setPE))).

Definition set_ptroot_spec : Action State :=
SAct_Action (StackCheck (getArgsCheck 1))
	           (fun s => 
                     (StackAct (pop +++ setRet 0%Z) +++ 
	             (MemAct (setPTROOT (nth 0 (StackGet getArgs s) 0%Z)))
                     ) s).

Definition hw_lib_spec : SpecHeap := fun l =>
  match l with
  | pe_enable_label => Some pe_enable_spec
  | set_ptroot_label => Some set_ptroot_spec
  | _ => None
  end.


Definition HW_kernel_init_spec : Action C_HW.State.t := {|
pre := (fun s => 
     C_HW.State.StackCheck (C_HW.State.Stack.getArgsCheck 0) s = true /\
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                      s.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                      s.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall pg, ValidPg pg = true -> s.(C_HW.State.memory).(HW_State.data) (PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(C_HW.State.memory).(HW_State.data) (PTROOT + pg * 8)= 0%Z) /\
      s.(C_HW.State.memory).(HW_State.pe) = true /\
      s.(C_HW.State.memory).(HW_State.ptroot) = PTROOT /\
      s.(C_HW.State.memory).(HW_State.tlb) = true
      );
rel := (fun s s' => False)
|}.

Definition HW_mem_init_spec : Action State :=
 {| pre := (fun (s : State) => StackCheck (getArgsCheck 0) s = true /\ s.(C_HW.State.memory).(HW_State.pe) = false); 
   rel := fun (s s' : State) => 
      (forall l,  ValidPA l = true -> ALE_State.PtDom l = true -> 
                      s'.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
     (forall l,  ValidPA l = true -> PE_State.PMMDom l = true -> 
                      s'.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
       (forall l : Z,
                ValidPA l = true ->
                (s'.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z \/
                s'.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 0%Z)
        ) /\
     s'.(C_HW.State.memory).(HW_State.pe) = false /\
     s'.(stack) = setRet 0 (pop s.(stack))
|}. 

Definition HW_pmap_init_spec : Action State :=
 {| pre := (fun (s : State) => StackCheck (getArgsCheck 0) s = true /\ 
                                           s.(C_HW.State.memory).(HW_State.pe) = false /\
			       (forall l : Z,
			                ValidPA l = true ->
			                (s.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z \/
			                s.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8)%Z = 0%Z)
			        ) /\
                                          (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                                                       s.(C_HW.State.memory).(HW_State.data) 
                                                                (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) /\
                                          (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                                                       s.(C_HW.State.memory).(HW_State.data)
                                                                (PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z)
                                           ); 
   rel := fun (s s' : State) => 
     (forall pg, ValidPg pg = true -> s'.(C_HW.State.memory).(HW_State.data) (PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s'.(C_HW.State.memory).(HW_State.data) (PTROOT + pg * 8)= 0%Z) /\
          (forall pg, ValidPg pg = true -> 
                           s'.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + pg * 8)%Z = 
                           s.(C_HW.State.memory).(HW_State.data) (PE_State.PMM_data_loc + pg * 8)%Z
           ) /\


     s'.(C_HW.State.memory).(HW_State.pe) = false /\
     s'.(stack) = setRet 0 (pop s.(stack))
|}. 


End HW_Lib.
