(* pe.v *)
(* Definition of the PE memory model *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import sact.
Require Import actsmp.

Require Import base_simp.



Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module PE_State <: Mem_T.

Definition PTROOT := PTROOT.

Record State := { data : Z -> Z; tlb : bool }.
Definition t := State.

Definition ValidVA (l :Z) : SCheck (State := State) := 
fun s => s.(tlb) &&
  let ppg := s.(data) (PTROOT + Pg(l) * 8) in
  ValidPA(ppg * PGSIZE + Off(l)).

Definition getMemCheck (l:Z) : SCheck (State := State) := 
fun s => ValidVA l s.
 
Definition getMem (l:Z) : SGet (State := State) Z :=
fun s => let ppg := s.(data) (PTROOT + Pg(l) * 8) in 
              let pa := (ppg * PGSIZE + Off l)%Z in
              s.(data) pa.

Definition setMemCheck (l : Z) : SCheck (State := State) := 
fun s => ValidVA l s.

Definition setMem (l z : Z) :  SAct (State := State) :=
fun s => let ppg := s.(data) (PTROOT + Pg(l) * 8) in 
              let pa := (ppg * PGSIZE + Off l)%Z in
              {| data := fun l' => if Zeq_bool pa l' then z else s.(data) l'; 
                      tlb := s.(tlb) |}.

Definition getTLB : SGet (State := State) bool :=
fun s => s.(tlb).

Definition setTLB (t : bool) :  SAct (State := State) :=
fun s =>  {| data := s.(data); tlb := t |}.




Definition PMM_data_loc : Z := (336 * PGSIZE)%Z. (* page 0x150 *)
Definition PMMsize := Zdiv_up NPAGES PGSIZE.

Definition PMMDom (l : Z) : bool := Zle_bool PMM_data_loc l  && Zlt_bool l (PMM_data_loc + NPAGES * 8)%Z.

Lemma PMMDom_zero : forall l, PMMDom l = PMMDom (Pg l * PGSIZE + Off 0).
intros.
unfold PMMDom, Pg.
Opaque PMM_data_loc.
Zeq_pull'; Zeq_norm; simpl in *; autosolve; elimtype False.
Transparent PMM_data_loc.

rewrite Off0 in *.
unfold PMM_data_loc in *.
replace (336 * 4096 + 32768)%Z with ((336 + 8) * 4096)%Z in H3; try omega.
replace (336 + 8)%Z with 344%Z in H3; try omega.
assert (l / 4096 >= 344)%Z.
omega.
clear H3.
assert (l >= 344 * 4096)%Z.
apply Zge_div_mult; try omega.
clear H H1 H0.
simpl in *; omega.

rewrite Off0 in *.
clear H0 H1.
unfold PMM_data_loc in *.
replace (336 * 4096 + 32768)%Z with ((336 + 8) * 4096)%Z in *; try omega.
replace (336 + 8)%Z with 344%Z in *; try omega.
assert (l / 4096 < 344)%Z.
omega.
clear H3.
assert ( l / 4096 >= 344)%Z.
apply Zge_mult_div_1; try omega.
omega.

rewrite Off0 in *.
clear H2.
unfold PMM_data_loc in *.
assert (336 <= l / 4096)%Z.
apply Zdiv_le_lower_bound; autosolve.
assert (336 > l / 4096)%Z; autosolve.

rewrite Off0 in *.
clear H3 H2.
unfold PMM_data_loc in *.
assert (336 <= l / 4096)%Z.
omega.
clear H1.
assert (336 * 4096 <= l)%Z.
apply Zle_div_mult; try omega.
apply Zgt_div_0 with PGSIZE; try omega.
assert (l / 4096 > 0)%Z; try omega.
Qed.


Lemma PMMDom_alloc : forall l, PMMDom l = true -> (100 <= Pg l < 512)%Z.
intros.
unfold PMMDom in H.
unfold Pg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold PMM_data_loc in *.
split.
apply Zle_mult_div; try omega.
replace (1409024)%Z with (344 * 4096)%Z in H2; try omega.
apply Zlt_mult_div_1; try omega.
Qed.

Lemma PMMDom_Loc : 
forall pg, ValidPg pg = true -> PE_State.PMMDom (PE_State.PMM_data_loc + pg * 8) = true.
intros.
unfold ValidPg, PMMDom in *.
Opaque PMM_data_loc.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
assert (pg * 8 >= 32768)%Z.
omega.
clear H4 H3.
unfold NPAGES in *.
autosolve.
Qed.

Lemma PMMDom_Valid : forall pg, ValidPg pg = true -> ValidPA (PE_State.PMM_data_loc + pg * 8) = true.
unfold ValidPg, ValidPA; intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold WordAligned.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Transparent PMM_data_loc.
unfold PMM_data_loc in *.
replace (336 * 4096)%Z with ((336 * 512) * 8)%Z in *; try omega.
rewrite math3 in H5; try omega.
rewrite mod_elim in H5; try omega.
elimtype False; apply H5; auto.

elimtype False.
clear H3.
unfold PMM_data_loc in H4.
unfold Pg in H4.
rewrite math1 in H4; try omega.
rewrite Zplus_comm in H4.
unfold NPAGES in *.
assert (pg * 8 / 4096 >= 4096 - 336)%Z.
omega.
clear H4.
replace (4096 - 336)%Z with (3760)%Z in H0; try omega.
assert (pg * 8 >= 3760 * 4096)%Z.
apply Zge_div_mult; try omega.
clear H0.
simpl in *.
replace (15400960)%Z with (1925120 * 8)%Z in H3; try omega.

elimtype False.
clear H4 H2.
unfold PMM_data_loc in *.
assert ((0- (336 * 4096)) > (pg * 8))%Z; try omega.

elimtype False.
clear H4 H2.
unfold PMM_data_loc in *.
assert ((0- (336 * 4096)) > (pg * 8))%Z; try omega.
Qed.

Lemma PMM_PT_separate : forall pg pg', 
ValidPg pg = true -> ValidPg pg' = true ->
Zeq_bool (PE_State.PMM_data_loc + pg * 8) (PE_State.PTROOT + pg' * 8) = false.
intros.
Opaque PMM_data_loc PTROOT.
unfold ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Transparent PMM_data_loc PTROOT.
unfold PMM_data_loc, PTROOT, common.PTROOT, NPAGES in *.
elimtype False.
omega.
Qed.

Opaque PTROOT PMM_data_loc.


Definition getPageInfoCheck (pg : Z) : SCheck (State := State) := fun s => ValidPg pg.

Definition getPageInfo (pg : Z) : SGet bool :=
  fun s => if Zeq_bool (s.(data) (PMM_data_loc + pg * 8)) 0%Z then false else true.

Definition setPageInfoCheck (pg : Z) : SCheck (State := State) := fun s => ValidPg pg.

Definition setPageInfo (pg :Z) (pi : bool) : SAct (State := State) := fun s => 
  {| data := fun l' => if Zeq_bool (PMM_data_loc + pg * 8) l' then (if pi then 1%Z else 0%Z) else s.(data) l'; 
                      tlb := s.(tlb) |}.

Definition DM : State -> Prop := fun s =>
   (forall pg, ValidPg pg = true -> PE_State.data s (PE_State.PTROOT + pg * 8) = pg).


(* getMem simplifications *)
Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, ValidVA, ValidPA in *; simpl; subst.
destruct s; simpl in *.
f_equal; simpl.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

End PE_State. 
