(* pm_hw.v *)
(* Memory relation between the PD and the HW memory models *)
(* Also includes proofs of compatibility for stubs of mem_init and pt_init in those two memory models *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.

Require Import mem_repr.
Require Import pm.
Require Import hw.

Module PM_HW <: Mem_Repr_T PM_State HW_State.

Opaque PTROOT.

Definition Repr_Mem (pm : PM_State.t) (hw : HW_State.t) :=
   (forall l, pm l = HW_State.data hw l) /\
   (HW_State.pe hw = false).

Lemma Repr_getMemCheck : forall (m1 : PM_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PM_State.getMemCheck l m1 = true -> HW_State.getMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold PM_State.getMemCheck, HW_State.getMemCheck in *.
unfold HW_State.ValidVA.
rewrite H1; simpl; autosolve.
Qed.

Lemma Repr_getMem : forall (m1 : PM_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, PM_State.getMemCheck l m1 = true -> PM_State.getMem l m1 = v -> HW_State.getMem l m2 = v.
intros.
red in H; repeat autoclean.
unfold PM_State.getMem, HW_State.getMem in *.
rewrite H2; simpl; autosolve.
Qed.

Lemma Repr_setMemCheck : forall (m1 : PM_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PM_State.setMemCheck l m1 = true -> HW_State.setMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold PM_State.setMemCheck, HW_State.setMemCheck in *.
unfold HW_State.ValidVA.
rewrite H1; simpl; autosolve.
Qed.

Lemma Repr_setMem : forall (m1 : PM_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , PM_State.setMemCheck l m1 = true -> Repr_Mem (PM_State.setMem l v m1) (HW_State.setMem l v m2).
intros.
red in H; repeat autoclean.
unfold PM_State.setMem, HW_State.setMem, Repr_Mem in *; simpl.
rewrite H1; simpl; repeat autoclean.
repeat split; autosolve.
intros; Zeq_pull'; Zeq_norm; autosolve.
Qed.
End PM_HW.

Require Import mem_sig.
Require Import mem_sig_ald.
Require Import hw_lib.

Require Import c.
Require Import pe.

Module C_PM_HW := C_Refine PM_State HW_State PM_HW C_PM C_HW.
Import PM_HW.

Opaque PTROOT PE_State.PTROOT PE_State.PMM_data_loc.

Lemma PM_HW_mem_exists : forall (s : HW_State.t), 
      s.(HW_State.pe) = false ->
(exists bs, PM_HW.Repr_Mem bs s).
intros.
destruct s; simpl in *.
exists data.
red; simpl.
repeat split; auto.
Qed.

Lemma mem_init_weak : ActionGTE HW_Lib.HW_mem_init_spec
            (C_PM_HW.RF.reprAction C_PM_HW.CRepr mem_sig_ald.PM_mem_init_spec).
split.

unfold HW_Lib.HW_mem_init_spec, mem_sig_ald.PM_mem_init_spec; simpl.
unfold C_HW.State.StackCheck, C_PM.State.StackCheck in *.
intros; repeat autoclean.
destruct s; simpl in *.
unfold C_PM_HW.CRepr; simpl.
unfold C_HW.State.Stack.getArgsCheck in *.
destruct stack; autosolve.
destruct f; autosolve.
destruct l; autosolve.
simpl in *; repeat autoclean.
destruct (PM_HW_mem_exists memory) as [bmem]; auto.
exists {| C_PM.State.stack := C_Stack.CallFrame nil :: stack; 
              C_PM.State.memory := bmem |}; simpl.
split; auto.

(* rel *)
unfold HW_Lib.HW_mem_init_spec, mem_sig_ald.PM_mem_init_spec; simpl.
intros; repeat autoclean.
unfold C_HW.State.StackCheck, C_PM.State.StackCheck in *.
destruct s; simpl in *.
destruct (PM_HW_mem_exists memory) as [bmem]; auto.
repeat autoclean.
destruct H0 with {| C_PM.State.stack := stack; 
              C_PM.State.memory := bmem |}; simpl in *; autosolve; repeat autoclean.
red; simpl; split; auto.

clear H0.
red in H3; repeat autoclean.
rewrite H0 in *; clear H0.
red in H3; repeat autoclean.
red in H2; repeat autoclean.
repeat split; autosolve.
intros.
rewrite <- H0.
apply H4; auto.
intros.
rewrite <- H0.
apply H5; auto.
intros.
rewrite <- H0; auto.
Qed.

Require Import pe.
Require Import ale.

Opaque PE_State.PTROOT ALE_State.PTROOT PE_State.PMM_data_loc.

Lemma PM_HW_pmap_exists : forall (s : HW_State.t), 
      s.(HW_State.pe) = false ->
(exists bs, PM_HW.Repr_Mem bs s).
intros.
destruct s; simpl in *.
exists data.
red; simpl.
repeat split; auto.
Qed.


Lemma pmap_init_weak : ActionGTE HW_Lib.HW_pmap_init_spec
            (C_PM_HW.RF.reprAction C_PM_HW.CRepr mem_sig_ald.PM_pmap_init_spec).
split.

unfold HW_Lib.HW_pmap_init_spec, mem_sig_ald.PM_pmap_init_spec; simpl.
unfold C_HW.State.StackCheck, C_PM.State.StackCheck in *.
intros; repeat autoclean.
destruct s; simpl in *.
unfold C_PM_HW.CRepr; simpl.
unfold C_HW.State.Stack.getArgsCheck in *.
destruct stack; autosolve.
destruct f; autosolve.
destruct l; autosolve.
simpl in *; repeat autoclean.
destruct (PM_HW_pmap_exists memory) as [bmem]; auto.
exists {| C_PM.State.stack := C_Stack.CallFrame nil :: stack; 
              C_PM.State.memory := bmem |}; simpl.
red in H; repeat autoclean.
repeat split; auto.

intros.
rewrite H.
destruct H1 with l; autosolve.

intros.
rewrite H.
rewrite H3; auto.

intros.
rewrite H.
rewrite H2; auto.

(* rel *)
unfold HW_Lib.HW_pmap_init_spec, mem_sig_ald.PM_pmap_init_spec; simpl.
intros; repeat autoclean.
unfold C_HW.State.StackCheck, C_PM.State.StackCheck in *.
destruct s; simpl in *.
destruct (PM_HW_mem_exists memory) as [bmem]; auto.
repeat autoclean.
specialize H0 with {| C_PM.State.stack := stack; 
              C_PM.State.memory := bmem |}; simpl in *; autosolve; repeat autoclean.
unfold C_PM_HW.CRepr in H0; simpl in *.
destruct H0; autosolve.
split; auto.
split; auto.
split; auto.

intros.
red in H5; repeat autoclean.
rewrite H5.
apply H2; autosolve.

split.
intros.
red in H5; repeat autoclean.
rewrite H5.
apply H4; autosolve.

intros.
red in H5; repeat autoclean.
rewrite H5.
apply H3; autosolve.

repeat autoclean.
rewrite <- H0.
red in H10; repeat autoclean.
repeat split; intros; try rewrite <- H10; autosolve.
apply H6; autosolve.
apply H7; auto.
red in H5; repeat autoclean.
rewrite <- H5; auto.
Qed.

(*********************************************************************)

Require Import label.

Definition ref := C_PM_HW.EqRef.

Lemma WF_PM_HW : forall CODE PSI,
         C_PM.SCAP.WFCode C_PM.State.expr_to_Prop C_PM.Mc_Machine C_PM.SCAP.SpecHeap.SpecHeap_empty CODE PSI ->

      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine 
                C_HW.SCAP.SpecHeap.SpecHeap_empty 
                (C_PM_HW.CMake CODE)
                (C_PM_HW.PsiMake PSI).
intros.
set (x := ref C_PM.SCAP.SpecHeap.SpecHeap_empty CODE PSI).
apply x; auto.
Qed.
