(* pmap_sig.v *)
(* The specifications of the pt_set and pt_lookup procedures in the ALE machine *)
(* Also includes the spec of the kernel_init stub in the ALE machine *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.

Require Import c.

Require Import sact.
Require Import actsmp.

Module pmap_sig.
Import C_ALE SCAP.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition pmap_set_spec : Action State := 
{| pre := (fun s => StackCheck (getArgsCheck 2) s = true /\
                let (vpage, entry) := (nth 0 (StackGet getArgs s) 0%Z, nth 1 (StackGet getArgs s) 0%Z) in
                ValidVPg vpage = true /\ ValidPg vpage = false /\ ValidPg entry = true /\ 
               DM s.(memory) /\ 
	       PtDom_allocated s.(memory) /\
	       MemGet getTLB s = true);
rel := (fun s s' => 
         let (vpage, entry) := (nth 0 (StackGet getArgs s) 0%Z, nth 1 (StackGet getArgs s) 0%Z) in
         s' = SAct_chain 
                   (StackAct (pop +++ setRet 0%Z)) 
                   (MemAct (setMem (PTROOT + vpage * 8) (entry))) s)
|}.

Definition pmap_lookup_spec : Action State := 
{| pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
               let vpage := nth 0 (StackGet getArgs s) 0%Z in
               ValidVPg vpage = true /\ ValidPg vpage = false /\ 
	       DM s.(memory) /\ 
	       PtDom_allocated s.(memory) /\
	       MemGet getTLB s = true);

rel := (fun s s' => 
            let vpage := nth 0 (StackGet getArgs s) 0%Z in
            let entry := MemGet (getMem (PTROOT + vpage * 8)) s in
            s' = StackAct (pop +++ setRet entry) s)
|}.

Definition pmap_spec : SpecHeap := fun l =>
  match l with
  | pmap_set_label => Some pmap_set_spec
  | pmap_lookup_label => Some pmap_lookup_spec
  | _ => None
  end.

Require Import ale. 
Require Import ale_lib.

Lemma lib_pmap_disj : C_ALE.SCAP.SpecHeap.SpecHeap_disjoint ALE_Lib.ale_lib_spec pmap_sig.pmap_spec.
red; intros.
destruct l; autosolve.
Qed.

(***********************************)
(* kernel_init stub *)

Definition ALE_kernel_init_spec : Action C_ALE.State.t := {|
pre := (fun s => 
     C_ALE.State.StackCheck (C_ALE.State.Stack.getArgsCheck 0) s = true /\
     (forall l, ValidPA l = true -> PtDom l = true -> 
                       s.(C_ALE.State.memory).(ALE_State.allocdata) (Pg l) = true) /\
     (forall pg, ValidPg pg = true -> s.(C_ALE.State.memory).(ALE_State.data) (PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(C_ALE.State.memory).(ALE_State.data) (PTROOT + pg * 8)= 0%Z) /\
      s.(C_ALE.State.memory).(ALE_State.tlb) = true
      );
rel := (fun s s' => False)
|}.


End pmap_sig.

