(* sact.v *)
(* SAct describes "simple actions", e.g. those that can be defined by a function *)
(* These simple actions can then be embedded as a complete action *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import spec.
Require Import scap.

Require Import zbool_assist.
Require Import Omega.
Require Import Logic.FunctionalExtensionality.

Definition SGet {State} A := State -> A %type.
Definition SCheck {State} := State -> bool.
Definition SAct {State} := (State -> State)%type.

Definition SAct_id {State} : SAct := fun (s : State) => s.

Definition SAct_Action {State} (c : SCheck) (f : SAct) :=
mkAct (fun (s : State) => c s = true)
      (fun (s : State) s' => f s = s').

Definition SCheck_and {State} (c c' : SCheck) : SCheck := fun (s:State) => c s && c' s.

Definition SCheck_chain {State} (c c' : SCheck) (f : SAct) : SCheck := fun (s:State) => c s && (c' (f s)).
Definition SAct_chain {State} (f f' : SAct) : SAct := fun (s : State) => f' (f s).
Notation "a +++ b" := (SAct_chain a b) (at level 80).


Definition SCheck_plus {State} (p : State -> bool) (c c' : SCheck)  : SCheck := fun (s : State) => if p s then c s else c' s.
Definition SAct_plus {State} (p : State -> bool) (f f' : SAct)  : SAct := fun (s : State) => if p s then f s else f' s.


Definition SCheck_Let {A : Type} {State} (get : SGet A) (ch : A -> SCheck) : SCheck :=
fun (s : State) => ch (get s) s.

Definition SAct_Let {A : Type} {State} (get : SGet A) (act : A -> SAct) : SAct :=
 fun (s : State) => act (get s) s. 

Definition SCheck_ensure {State} (get : SCheck) (f : SCheck) : SCheck :=
  fun (s : State) => get s && f s.

(* Properties *)

Lemma ActionChain_SAct {State : Type} : forall c c' a a', 
  ActionGTE 
    (SAct_Action (State := State) (SCheck_chain c c' a) (SAct_chain a a'))
    (ActionChain (SAct_Action c a) (SAct_Action c' a')).
intros.
unfold SAct_Action, SAct_chain, SCheck_chain.
firstorder; simpl in *.
destruct (c s); auto.
subst.
destruct (c' (a s)); destruct (c s); auto.
subst; auto.
Qed.

Lemma ActionChain_SAct' {State} : forall c c' a a', 
  ActionGTE 
    (ActionChain (SAct_Action c a) (SAct_Action c' a'))
    (SAct_Action (State := State) (SCheck_chain c c' a) (SAct_chain a a')).
intros.
unfold SAct_Action, SAct_chain, SCheck_chain.
firstorder; simpl in *.
specialize (H0 (a s)).
firstorder.
Qed.


Lemma pre_SActAction : forall State c a (s : State),
  pre (SAct_Action c a) s = (c s = true).
firstorder.
Qed.
Hint Rewrite pre_SActAction : actsmp.

Lemma rel_SActAction : forall State c a (s s' : State),
  rel (SAct_Action c a) s s' = (a s = s'). 
firstorder.
Qed.
Hint Rewrite rel_SActAction : actsmp.

Ltac sact_structured := 
progress (
try unfold SAct_Let, SAct_chain, SCheck_chain, SCheck_Let, SCheck_and, SAct_plus, SCheck_plus in *;
repeat autorewrite with actsmp in *).

Ltac sact_auf :=
repeat (
try unfold SAct_Action, SAct_Let, SAct_chain, SCheck_chain, SCheck_Let, SCheck_and, SAct_plus, SCheck_plus in *;
simpl in *).
