(* as_release.v *)
(* Defines and certifies the code of the as_release procedure *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.


Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import pmap.
Require Import pmap_lib.
Require Import mach.as_sig.

Require Import label.

Module as_release.
Import C_PMAP State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.


(***************************************************)

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Axiom pmap_lookup_flabel : FLABEL.
Axiom pmap_set_flabel : FLABEL.
Axiom mem_free_flabel : FLABEL.

Definition as_release : Process :=
  Iperf (perf_readargs (page_var :: nil)) :::
  Iperf (perf_fcall pmap_lookup_flabel (expr_var page_var :: nil)) :::
  Icall pmap_lookup_label :::
  Iperf (perf_readret entry_var) :::
  Iperf (perf_fcall mem_free_flabel (expr_var entry_var :: nil)) :::
  Icall mem_free_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_fcall pmap_set_flabel (expr_var page_var :: expr_z 0%Z :: nil)) :::
  Icall pmap_set_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_ret (expr_z 0%Z)).


Definition as_release_spec' : Action State :=
  Actions (perf_readargs (page_var :: nil)) ;;
  Actions (perf_fcall pmap_lookup_flabel (expr_var page_var :: nil)) ;;
  PMAP_Lib.pmap_lookup_spec ;;
  Actions (perf_readret entry_var) ;;
  Actions (perf_fcall mem_free_flabel (expr_var entry_var :: nil)) ;;
  PMAP_Lib.mem_free_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_fcall pmap_set_flabel (expr_var page_var :: expr_z 0%Z :: nil)) ;;
  PMAP_Lib.pmap_set_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_ret (expr_z 0%Z)).

Definition WF_as_release' : WFProc expr_to_Prop Mc_Machine PMAP_Lib.pmap_lib_spec as_release_spec' as_release.
unfold as_release_spec'; unfold as_release; auto 999.
Qed.


Definition WF_as_release : WFProc expr_to_Prop Mc_Machine PMAP_Lib.pmap_lib_spec as_sig.as_release_spec as_release.
apply WFProc_weaken with as_release_spec'; [|apply WF_as_release'].
unfold as_sig.as_release_spec, as_release_spec'.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.

split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

repeat autoclean.
destruct s; simpl in *; autosolve.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
unfold pushArgs, StackCheck, getArgsCheck, pop; simpl in *.
repeat autoclean.

repeat split; simpl in *; repeat autoclean; simpl in *; autosolve.
rewrite H0; rewrite H1; autosolve.
rewrite H4; simpl in *; autosolve.
rewrite H4; simpl in *; autosolve.
rewrite H4; simpl in *; autosolve.
unfold ValidPg in H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H4; simpl in *; autosolve.
unfold ValidPg in H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H4; simpl in *; autosolve.
rewrite H4 in *; simpl in *; autosolve.
unfold PMAP_Lib.freepg, PMAP_Lib.PMAP_free, StackRets in *; simpl in *.
decompose record H6; clear H6.
rewrite H7 in *; clear H7; simpl in *; auto.
rewrite H4 in *; simpl in *; autosolve.
unfold PMAP_Lib.freepg, PMAP_Lib.PMAP_free, StackRets in *; simpl in *.
decompose record H7; clear H7.
rewrite H6 in *; clear H6; simpl in *; auto.

rewrite H4 in *; simpl in *; autosolve.
unfold PMAP_Lib.freepg, PMAP_Lib.PMAP_free, StackRets in *; simpl in *.
decompose record H8; clear H8.
rewrite H6 in *; clear H6; simpl in *; auto.
rewrite H0; rewrite H1; simpl in *; auto.

rewrite H4 in *; simpl in *; autosolve.
unfold PMAP_Lib.freepg, PMAP_Lib.PMAP_free, StackRets in *; simpl in *.
decompose record H10; clear H10.
rewrite H6 in *; clear H6; simpl in *; auto.

(***********************************)
(* rel *)
(***********************************)

repeat autoclean; simpl in *.
rewrite H15 in *; simpl in *.
unfold PMAP_Lib.freepg, PMAP_Lib.PMAP_free, StackRets in *; simpl in *.
decompose record H9; clear H9.
rewrite H4 in *; clear H4; simpl in *.
unfold as_sig.vfree, as_sig.PMAP_vfree; simpl in *.
repeat split; autosolve.
rewrite H7; auto.
rewrite H15 in *; simpl in *.
rewrite H0; auto.
Qed.

End as_release.