(* as_reserve.v *)
(* Defined and certified the code of the as_reserve procedure *)
(* due to small inconsistencies between the paper and the code, this is also the as_request proc *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.


Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import pmap.
Require Import pmap_lib.
Require Import mach.as_sig.

Require Import label.

Module as_reserve.
Import C_PMAP.
Import C_PMAP.State.
Import C_PMAP.SCAP.

Require Import sact.
Require Import actsmp.


(***************************************************)

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition ife : Expr.expr_nomem := (expr_nomem_binop EQ (expr_nomem_var entry_var) (expr_nomem_z 0%Z)).

Axiom pmap_lookup_flabel : FLABEL.
Axiom pmap_set_flabel : FLABEL.
Axiom mem_free_flabel : FLABEL.
Axiom mem_alloc_flabel : FLABEL.

Definition as_reserve : C_PMAP.SCAP.Process.Process :=
  Iperf (perf_readargs (page_var :: nil)) :::
  Iperf (perf_fcall mem_alloc_flabel nil) :::
  Icall mem_alloc_label :::
  Iperf (perf_readret entry_var) :::
  Iplus ife
  	(Iperf (perf_ret (expr_z 0%Z)))

	(
  	Iperf (perf_fcall pmap_set_flabel (expr_var page_var :: expr_var entry_var :: nil)) :::
	Icall pmap_set_label :::
	Iperf (perf_readret junk_var) :::
	Iperf (perf_ret (expr_var page_var))
	).


Definition as_reserve_spec' : Action State :=
  Actions (perf_readargs (page_var :: nil)) ;;
  Actions (perf_fcall mem_alloc_flabel nil) ;;
  PMAP_Lib.mem_alloc_spec ;;
  Actions (perf_readret entry_var) ;;
  ActionPlus (expr_to_Prop ife)
  	(Actions (perf_ret (expr_z 0%Z)))

	(
  	Actions (perf_fcall pmap_set_flabel (expr_var page_var :: expr_var entry_var :: nil)) ;;
	PMAP_Lib.pmap_set_spec ;;
	Actions (perf_readret junk_var) ;;
	Actions (perf_ret (expr_var page_var))
	).

Definition WF_as_reserve' : WFProc C_PMAP.State.expr_to_Prop Mc_Machine 
                               PMAP_Lib.pmap_lib_spec as_reserve_spec' as_reserve.
unfold as_reserve_spec'; unfold as_reserve; auto 999.
Qed.

Transparent pushArgs setVar getArgs pushVar pop.

Definition WF_as_reserve : WFProc C_PMAP.State.expr_to_Prop Mc_Machine 
      PMAP_Lib.pmap_lib_spec as_sig.as_reserve_spec as_reserve.
apply WFProc_weaken with as_reserve_spec'; [|apply WF_as_reserve'].
unfold as_sig.as_reserve_spec, as_reserve_spec', ife, expr_to_Prop.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

unfold StackCheck.
repeat autoclean.
repeat split; autosolve; repeat autoclean; simpl in *; autosolve.

unfold PMAP_Lib.allocpg, PMAP_Lib.PMAP_alloc, PMAP_Lib.allocfail in *; simpl in *.
destruct H3; simpl in *.
destruct H2; simpl in *.
decompose record H2; clear H2.
unfold StackRets in *.
rewrite H8; simpl in *; auto.
subst; simpl in *; auto.

unfold PMAP_Lib.allocpg, PMAP_Lib.PMAP_alloc, PMAP_Lib.allocfail in *; simpl in *.
destruct H4.
destruct H2.
decompose record H2; clear H2; simpl in *.
unfold StackRets in *.
rewrite H8 in *; simpl in *; clear H8.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
right.
repeat split; autosolve.
intros; subst; simpl in *; autosolve.
rewrite H0; rewrite H1; simpl.
rewrite H4.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros; subst; simpl in *; autosolve.
destruct H2.
destruct H2; subst; simpl in *; autosolve.
intros.
destruct H2.
destruct H2; subst; simpl in *.
destruct H2; subst; simpl in *.
destruct H2; subst; simpl in *; autosolve.
subst; simpl in *.
left; repeat split; autosolve.

(***********************************)
(* rel *)
(***********************************)

repeat autoclean.
simpl in *.

unfold getVarCheck, setVar, getVar, getArgsCheck in *; simpl in *.

destruct s; simpl in *; autosolve.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
unfold as_sig.vallocpg, PMAP_Lib.allocpg,
      StackRets, as_sig.PMAP_valloc in *; simpl in *.
destruct H6; repeat autoclean; simpl in *.
(* allocated *)
left; exists x.
rewrite H6 in *; clear H6; simpl in *.
clear H3.

assert (~(true = true /\ (if Zeq_bool x 0 then 1%Z else 0%Z) <> 0%Z)).
red; intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

specialize (H4 H3); clear H3.
repeat autoclean; simpl in *.
red in H5; decompose record H5; clear H5.
repeat split; autosolve.
rewrite H8; auto.
makenew' (Zeq_bool x 0%Z); Zeq_norm; simpl in *; autosolve.
(* not allocated *)
clear H4; simpl in *.
right.
rewrite <- H3; autosolve.
split; autosolve.
Qed.

End as_reserve.