(* mem_free.v *)
(* Definition and certification of the code of mem_free procedure *)
Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.


Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import pe.
Require Import mach.mem_sig.

Module mem_free.
Import C_PE State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.

Opaque PTROOT PE_State.PTROOT PE_State.PMM_data_loc.


(***************************************************)

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | _ => None
   end.

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).
Definition PMM_expr (pg : expr) : expr :=
  (expr_binop PLUS (expr_z PMM_data_loc) (expr_binop TIMES pg (expr_z 8%Z))).

Definition mem_free : Process :=
  Iperf (perf_readargs (page_var :: nil)) :::
  Iperf (perf_store (PMM_expr (expr_var page_var)) (expr_z 0%Z)) :::
  Iperf (perf_ret (expr_z 0%Z)).

Definition mem_free_spec' : Action State :=
  Actions (perf_readargs (page_var :: nil)) ;;
  Actions (perf_store (PMM_expr (expr_var page_var)) (expr_z 0%Z)) ;;
  Actions (perf_ret (expr_z 0%Z)).

Definition WF_mem_free' : WFProc expr_to_Prop Mc_Machine Psi mem_free_spec' mem_free.
unfold mem_free_spec'; unfold mem_free; auto 999.
Qed.

Definition WF_mem_free : WFProc expr_to_Prop Mc_Machine Psi mem_sig.mem_free_spec mem_free.
apply WFProc_weaken with mem_free_spec'; [|apply WF_mem_free'].
unfold mem_sig.mem_free_spec, mem_free_spec'. 

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.

split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

(***************)
repeat autoclean.
repeat split; autosolve.
intros; subst; simpl in *.
destruct s; simpl in *.
makenew (nth 0 (getArgs stack0) 0%Z).
unfold setMemCheck.
unfold ValidVA.
unfold getTLB in H2.
rewrite H2; simpl.
rewrite H3.
rewrite <- Addr_parts.
apply PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply ValidPA_ValidPg.
apply PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
intros; repeat autoclean; subst; simpl in *; auto.

(* rel *)
repeat autoclean; simpl in *.
rewrite H2.
f_equal.
unfold setPageInfo, setMem.
rewrite H4.
rewrite <- Addr_parts; autosolve.
apply ValidPA_ValidPg.
apply PMMDom_Valid.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

End mem_free.
