(* pmap_set.v *)
(* Defines and certifies the code of the pt_set procedure *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.

Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import ale.
Require Import mach.pmap_sig.

Module pmap_set.
Import C_ALE State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.

Opaque PTROOT ALE_State.PTROOT.
(***************************************************)

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | _ => None
   end.

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition pmap_set : Process :=
  Iperf (perf_readargs (page_var :: entry_var :: nil)) :::
  Iperf (perf_store  (expr_binop PLUS (expr_z PTROOT) 
                                   (expr_binop TIMES (expr_var page_var) (expr_z 8%Z)))
                              (expr_var entry_var)
           ) :::
  Iperf (perf_ret (expr_z 0%Z)).

Definition pmap_set_spec' : Action State :=
  Actions (perf_readargs (page_var :: entry_var :: nil)) ;;
  Actions (perf_store  (expr_binop PLUS (expr_z PTROOT) 
                                   (expr_binop TIMES (expr_var page_var) (expr_z 8%Z)))
                              (expr_var entry_var)
           ) ;;
  Actions (perf_ret (expr_z 0%Z)).


Definition WF_pmap_set' : WFProc expr_to_Prop Mc_Machine Psi pmap_set_spec' pmap_set.
unfold pmap_set_spec'; unfold pmap_set; auto 999.
Qed.

Definition WF_pmap_set : WFProc expr_to_Prop Mc_Machine Psi pmap_sig.pmap_set_spec pmap_set.
apply WFProc_weaken with pmap_set_spec'; [|apply WF_pmap_set'].
unfold pmap_sig.pmap_set_spec, pmap_set_spec'.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.

split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).
repeat split; intros; repeat autoclean; subst; simpl in *; autosolve.
unfold ALE_State.setMemCheck.
unfold vaCheck, transCheck, paCheck, isPhys.
assert (ValidPA (PTROOT + nth 0 (getArgs (stack s)) 0%Z * 8) = true).
apply PtDom_Valid; auto.
assert (ValidVA (PTROOT + nth 0 (getArgs (stack s)) 0%Z * 8) = true).
apply ValidPA_ValidVA; auto.
rewrite H4; simpl; auto.
rewrite H7; rewrite H0; simpl.
unfold getTLB in H6; rewrite H6; simpl.
rewrite ValidPA_ValidPg; auto.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H5; simpl; auto.
apply PtDom_PTROOT; auto.

(* rel *)
repeat autoclean; simpl in *; auto.
Qed.

End pmap_set.
