(* zbool_assist.v *)
(* Tactics and Axioms for converting boolean Z comparisons into Propositions *)
(* Also includes the makenew tactics that are used throughout *)

Require Import ZArith.
Require Import Bool.

Ltac save x :=
let blah := fresh in 
let newdecl := fresh in
let neweq := fresh in
assert (exists a, x = a) as blah; [eexists; auto |]; 
inversion_clear blah as [newdecl neweq].

Ltac saveas x newdecl neweq :=
let blah := fresh in
assert (exists a, x = a) as blah; [exists x; auto |]; 
inversion_clear blah as [newdecl neweq].

Ltac saveasl x newdecl neweq :=
let blah := fresh in
assert (exists a, a = x) as blah; [exists x; auto |]; 
inversion_clear blah as [newdecl neweq].


Ltac makenew a :=
        let newdecl := fresh in
        let neweq := fresh in
        saveas a newdecl neweq ; 
        rewrite neweq in * |- * 
        (*;  destruct newdecl; try clear a neweq*)
.

Ltac makenew' a :=
        let newdecl := fresh in
        let neweq := fresh in
        saveas a newdecl neweq ; 
        rewrite neweq in * |- * ; 
        destruct newdecl; try clear a neweq
.
Ltac makenew'l a :=
        let newdecl := fresh in
        let neweq := fresh in
        saveasl a newdecl neweq ; 
        rewrite <- neweq in * |- * ; 
        destruct newdecl; try clear a neweq
.


Lemma Zeq_bool_refl : forall a, Zeq_bool a a = true.
induction a; auto.
induction p; auto.
induction p; auto.
Qed.

Lemma Zlt_bool_lt : forall a b, Zlt_bool a b = true -> (a < b)%Z.
intros.
unfold Zlt_bool in *.
unfold Zlt.
destruct Zcompare in *; try inversion H; auto.
Qed.
Implicit Arguments Zlt_bool_lt [a b].

Lemma Zle_bool_le : forall a b, Zle_bool a b = true -> (a <= b)%Z.
intros.
unfold Zle_bool in *.
unfold Zle.
destruct Zcompare in *; try inversion H; auto.
red; intros; inversion H0.
red; intros; inversion H0.
Qed.
Implicit Arguments Zle_bool_le [a b].

Lemma Zgt_bool_gt : forall a b, Zgt_bool a b = true -> (a > b)%Z.
intros.
unfold Zgt_bool in *.
unfold Zgt.
destruct Zcompare in *; try inversion H; auto.
Qed.
Implicit Arguments Zgt_bool_gt [a b].

Lemma Zge_bool_ge : forall a b, Zge_bool a b = true -> (a >= b)%Z.
intros.
unfold Zge_bool in *.
unfold Zge.
destruct Zcompare in *; try inversion H; auto.
red; intros; inversion H0.
red; intros; inversion H0.
Qed.
Implicit Arguments Zge_bool_ge [a b].

Lemma Zeq_bool_eq : forall a b, Zeq_bool a b = true -> a = b.
intros.
unfold Zeq_bool in *.
destruct (Z_dec a b).
destruct s; auto.
unfold Zlt in *.
rewrite z in H; inversion H.
unfold Zgt in *.
rewrite z in H; inversion H.
auto.
Qed.
Implicit Arguments Zeq_bool_eq [a b].

Lemma Zneq_bool_neq : forall a b, Zneq_bool a b = true -> a <> b.
intros.
red; intros.
subst.
unfold Zneq_bool in *.
rewrite Zcompare_refl in H.
inversion H.
Qed.
Implicit Arguments Zneq_bool_neq [a b].

Lemma Zlt_bool_false_ge : forall a b, Zlt_bool a b = false -> (a >= b)%Z.
intros.
unfold Zlt_bool, Zge in *.
red; intros.
rewrite H0 in H; inversion H.
Qed.
Implicit Arguments Zlt_bool_false_ge [a b].

Lemma Zle_bool_false_gt : forall a b, Zle_bool a b = false -> (a > b)%Z.
intros.
unfold Zle_bool, Zgt in *.
destruct (Zcompare a b); try inversion H; auto.
Qed.
Implicit Arguments Zle_bool_false_gt [a b].

Lemma Zgt_bool_false_le : forall a b, Zgt_bool a b = false -> (a <= b)%Z.
intros.
unfold Zgt_bool, Zle in *.
red; intros.
rewrite H0 in H; try inversion H; auto.
Qed.
Implicit Arguments Zgt_bool_false_le [a b].

Lemma Zge_bool_false_lt : forall a b, Zge_bool a b = false -> (a < b)%Z.
intros.
unfold Zge_bool, Zlt in *.
destruct (Zcompare a b); try inversion H; auto.
Qed.
Implicit Arguments Zge_bool_false_lt [a b].

Lemma Zeq_bool_false_neq : forall a b, Zeq_bool a b = false -> a <> b.
intros.
red; intros.
subst.
unfold Zeq_bool in *; simpl in *.
rewrite Zcompare_refl in H.
inversion H.
Qed.
Implicit Arguments Zeq_bool_false_neq [a b].

Lemma Zneq_bool_false_eq : forall a b, Zneq_bool a b = false -> a = b.
intros.
unfold Zneq_bool in *.
apply Zcompare_Eq_eq.
destruct (Zcompare a b); try inversion H; auto.
Qed.
Implicit Arguments Zneq_bool_false_eq [a b].

Lemma Zlt_lt_bool : forall a b, (a < b)%Z -> Zlt_bool a b = true.
intros.
unfold Zlt_bool, Zlt in *.
rewrite H; auto.
Qed.
Implicit Arguments Zlt_lt_bool [a b].

Lemma Zle_le_bool : forall a b, (a <= b)%Z -> Zle_bool a b = true.
intros.
unfold Zle_bool, Zle in *.
destruct (Zcompare a b); auto.
Qed.
Implicit Arguments Zle_le_bool [a b].

Lemma Zgt_gt_bool : forall a b, (a > b)%Z -> Zgt_bool a b = true.
intros.
unfold Zgt_bool, Zgt in *.
rewrite H; auto.
Qed.
Implicit Arguments Zgt_gt_bool [a b].

Lemma Zge_ge_bool : forall a b, (a >= b)%Z -> Zge_bool a b = true.
intros.
unfold Zge_bool, Zge in *.
destruct (Zcompare a b); auto.
Qed.
Implicit Arguments Zge_ge_bool [a b].

Lemma Zeq_eq_bool : forall a b, a = b -> Zeq_bool a b = true.
intros.
unfold Zeq_bool; subst; auto.
rewrite Zcompare_refl; auto.
Qed.
Implicit Arguments Zeq_eq_bool [a b].

Lemma Zneq_neq_bool : forall a b, a <> b -> Zneq_bool a b = true.
intros.
unfold Zneq_bool.
makenew (Zcompare a b).
destruct H0; auto.
rewrite (Zcompare_Eq_eq a b) in H; auto.
Qed.
Implicit Arguments Zneq_neq_bool [a b].

Lemma Zge_lt_bool_false : forall a b, (a >= b)%Z -> Zlt_bool a b = false.
intros.
unfold Zge, Zlt_bool in *.
makenew (Zcompare a b).
destruct H0; auto.
elimtype False; apply H; auto.
Qed.
Implicit Arguments Zge_lt_bool_false [a b].

Lemma Zgt_le_bool_false : forall a b, (a > b)%Z -> Zle_bool a b = false.
intros.
unfold Zgt, Zle_bool in *.
rewrite H; auto.
Qed.
Implicit Arguments Zgt_le_bool_false [a b].

Lemma Zle_gt_bool_false : forall a b, (a <= b)%Z -> Zgt_bool a b = false.
intros.
unfold Zle, Zgt_bool in *.
makenew (Zcompare a b).
destruct H0; auto.
elimtype False; apply H; auto.
Qed.
Implicit Arguments Zle_gt_bool_false [a b].

Lemma Zlt_ge_bool_false : forall a b, (a < b)%Z -> Zge_bool a b = false.
intros.
unfold Zlt, Zge_bool in *.
rewrite H; auto.
Qed.
Implicit Arguments Zlt_ge_bool_false [a b].

Lemma Zneq_eq_bool_false : forall a b, a <> b -> Zeq_bool a b = false.
intros.
unfold Zeq_bool.
makenew (Zcompare a b).
destruct H0; auto.
rewrite (Zcompare_Eq_eq a b) in H; auto.
elimtype False; apply H; auto.
Qed.
Implicit Arguments Zneq_eq_bool_false [a b].

Lemma Zeq_neq_bool_false : forall a b, a = b -> Zneq_bool a b = false.
intros; subst.
unfold Zneq_bool.
rewrite Zcompare_refl; auto.
Qed.
Implicit Arguments Zeq_neq_bool_false [a b].

Lemma Zcmp_eq_eq : forall a b, (a ?= b)%Z = Eq -> (a = b)%Z.
exact Zcompare_Eq_eq.
Qed.

Lemma Zcmp_lt_lt : forall a b, (a ?= b)%Z = Lt -> (a < b)%Z.
auto.
Qed.

Lemma Zcmp_gt_gt : forall a b, (a ?= b)%Z = Gt -> (a > b)%Z.
auto.
Qed.

Ltac imp_rewrite h imp :=
  let res := fresh in
  set (res := imp _ _ h); generalize res; clear res h; intros h.

Ltac Zeq_norm :=
repeat
match goal with
 | h1 : _ = Zeq_bool ?a ?b |- _ => symmetry in h1
 | h1 : _ = Zneq_bool ?a ?b |- _ => symmetry in h1
 | h1 : _ = Zlt_bool ?a ?b |- _ => symmetry in h1
 | h1 : _ = Zle_bool ?a ?b |- _ => symmetry in h1
 | |- _ = Zeq_bool ?a ?b => symmetry
 | |- _ = Zneq_bool ?a ?b => symmetry
 | |- _ = Zlt_bool ?a ?b => symmetry
 | |- _ = Zle_bool ?a ?b => symmetry
end;
repeat
match goal with
 | h1 : (?a ?= ?b)%Z = Lt |- _ => generalize (Zcmp_lt_lt a b h1); clear h1; intros h1
 | h1 : (?a ?= ?b)%Z = Gt |- _ => generalize (Zcmp_gt_gt a b h1); clear h1; intros h1
 | h1 : (?a ?= ?b)%Z = Eq |- _ => generalize (Zcmp_eq_eq a b h1); clear h1; intros h1
end;
repeat
let xxx1 := fresh in 
let xxx2 := fresh in
match goal with
 | h1 : Zeq_bool ?a ?b = true |- _ => imp_rewrite h1 Zeq_bool_eq
 | h1 : Zneq_bool ?a ?b = true |- _ => imp_rewrite h1 Zneq_bool_neq
 | h1 : Zlt_bool ?a ?b = true |- _ => imp_rewrite h1 Zlt_bool_lt
 | h1 : Zle_bool ?a ?b = true |- _ => imp_rewrite h1 Zle_bool_le
 | h1 : Zgt_bool ?a ?b = true |- _ => imp_rewrite h1 Zgt_bool_gt
 | h1 : Zge_bool ?a ?b = true |- _ => imp_rewrite h1 Zge_bool_ge
 | h1 : Zeq_bool ?a ?b = false |- _ => imp_rewrite h1 Zeq_bool_false_neq
 | h1 : Zneq_bool ?a ?b = false |- _ => imp_rewrite h1 Zneq_bool_false_eq
 | h1 : Zlt_bool ?a ?b = false |- _ => imp_rewrite h1 Zlt_bool_false_ge
 | h1 : Zle_bool ?a ?b = false |- _ => imp_rewrite h1 Zle_bool_false_gt
 | h1 : Zgt_bool ?a ?b = false |- _ => imp_rewrite h1 Zgt_bool_false_le
 | h1 : Zge_bool ?a ?b = false |- _ => imp_rewrite h1 Zge_bool_false_lt

 | |- Zeq_bool ?a ?b = true => apply Zeq_eq_bool
 | |- Zneq_bool ?a ?b = true => apply Zneq_neq_bool
 | |- Zlt_bool ?a ?b = true => apply Zlt_lt_bool
 | |- Zle_bool ?a ?b = true  => apply Zle_le_bool
 | |- Zgt_bool ?a ?b = true => apply Zgt_gt_bool
 | |- Zge_bool ?a ?b = true  => apply Zge_ge_bool
 | |- Zeq_bool ?a ?b = false => apply Zneq_eq_bool_false
 | |- Zneq_bool ?a ?b = false => apply Zeq_neq_bool_false
 | |- Zlt_bool ?a ?b = false => apply Zge_lt_bool_false
 | |- Zle_bool ?a ?b = false => apply Zgt_le_bool_false
 | |- Zgt_bool ?a ?b = false => apply Zle_gt_bool_false
 | |- Zge_bool ?a ?b = false => apply Zlt_ge_bool_false
end.

Ltac Zeq_pull :=
repeat
match goal with
| h1 : context [Zeq_bool ?a ?b] |- _ => 
     lazymatch type of h1 with 
        | Zeq_bool _ _ = _ => fail
        | _ => makenew (Zeq_bool a b)
     end
| h1 : context [Zneq_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zneq_bool _ _ = _ => fail
        | _ => makenew (Zneq_bool a b)
     end
| h1 : context [Zle_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zle_bool _ _ = _ => fail
        | _ => makenew (Zle_bool a b)
     end
| h1 : context [Zlt_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zlt_bool _ _ = _ => fail
        | _ => makenew (Zlt_bool a b)
     end
| h1 : context [Zge_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zge_bool _ _ = _ => fail
        | _ => makenew (Zge_bool a b)
     end
| h1 : context [Zgt_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zgt_bool _ _ = _ => fail
        | _ => makenew (Zgt_bool a b)
     end
| |- context [Zeq_bool ?a ?b ] => makenew (Zeq_bool a b)
| |- context [Zneq_bool ?a ?b ] => makenew (Zneq_bool a b)
| |- context [Zle_bool ?a ?b ] => makenew (Zle_bool a b)
| |- context [Zlt_bool ?a ?b ] => makenew (Zlt_bool a b)
| |- context [Zge_bool ?a ?b ] => makenew (Zge_bool a b)
| |- context [Zgt_bool ?a ?b ] => makenew (Zgt_bool a b)
end.


Ltac Zeq_pull' :=
repeat
match goal with
| h1 : context [Zeq_bool ?a ?b] |- _ => 
     lazymatch type of h1 with 
        | Zeq_bool _ _ = _ => fail
        | _ => makenew' (Zeq_bool a b)
     end
| h1 : context [Zneq_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zneq_bool _ _ = _ => fail
        | _ => makenew' (Zneq_bool a b)
     end
| h1 : context [Zle_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zle_bool _ _ = _ => fail
        | _ => makenew' (Zle_bool a b)
     end
| h1 : context [Zlt_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zlt_bool _ _ = _ => fail
        | _ => makenew' (Zlt_bool a b)
     end
| h1 : context [Zge_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zge_bool _ _ = _ => fail
        | _ => makenew' (Zge_bool a b)
     end
| h1 : context [Zgt_bool ?a ?b] |- _ => 
     lazymatch type of h1 with
        | Zgt_bool _ _ = _ => fail
        | _ => makenew' (Zgt_bool a b)
     end
| |- context [Zeq_bool ?a ?b ] => makenew' (Zeq_bool a b)
| |- context [Zneq_bool ?a ?b ] => makenew' (Zneq_bool a b)
| |- context [Zle_bool ?a ?b ] => makenew' (Zle_bool a b)
| |- context [Zlt_bool ?a ?b ] => makenew' (Zlt_bool a b)
| |- context [Zge_bool ?a ?b ] => makenew' (Zge_bool a b)
| |- context [Zgt_bool ?a ?b ] => makenew' (Zgt_bool a b)
end.

